@@ -1,7 +1,7 @@
 /*
   Copyright (C) 2002, 2004, 2005, 2006, 2007 Jeroen Frijters
   Copyright (C) 2006 Active Endpoints, Inc.
-  Copyright (C) 2006, 2007 Volker Berlin
+  Copyright (C) 2006, 2007, 2008 Volker Berlin (i-net software)
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -41,7 +41,7 @@ namespace ikvm.awt
         private readonly Bitmap bitmap;
 
         internal BitmapGraphics(Bitmap bitmap)
-            : base(Graphics.FromImage(bitmap), null, Color.White)
+            : base(Graphics.FromImage(bitmap), null, Color.White, Color.Black)
         {
             this.bitmap = bitmap;
         }
@@ -63,7 +63,7 @@ namespace ikvm.awt
         private readonly Control control;
 
         internal ComponentGraphics(NetComponentPeer peer)
-            : base(peer.control.CreateGraphics(), peer.component.getFont(), peer.control.BackColor)
+            : base(peer.control.CreateGraphics(), peer.component.getFont(), peer.control.ForeColor, peer.control.BackColor)
         {
             control = peer.control;
         }
@@ -86,7 +86,7 @@ namespace ikvm.awt
         protected Graphics g;
         private java.awt.Color javaColor;
         private java.awt.Paint javaPaint;
-        private Color color = SystemColors.WindowText;
+        private Color color;
         private Color bgcolor;
         private java.awt.Font font;
         private java.awt.Stroke stroke;
@@ -94,8 +94,9 @@ namespace ikvm.awt
         private Font netfont;
         private Brush brush;
         private Pen pen;
+        private java.awt.Composite composite = java.awt.AlphaComposite.SrcOver;
 
-        protected NetGraphics(Graphics g, java.awt.Font font, Color bgcolor)
+        protected NetGraphics(Graphics g, java.awt.Font font, Color fgcolor, Color bgcolor)
         {
             if (font == null)
             {
@@ -103,6 +104,7 @@ namespace ikvm.awt
             }
             this.font = font;
             netfont = ((NetFontPeer)font.getPeer()).netFont;
+			this.color = fgcolor;
             this.bgcolor = bgcolor;
             init(g);
         }
@@ -344,21 +346,43 @@ namespace ikvm.awt
                 g.DrawPath(pen, gp);
         }
 
-        public override void drawString(java.text.AttributedCharacterIterator param1, int param2, int param3)
+        public override void drawString(java.text.AttributedCharacterIterator iterator, int x, int y)
         {
-            throw new NotImplementedException();
+            if (iterator == null) {
+                throw new java.lang.NullPointerException("AttributedCharacterIterator is null");
+            }
+            if (iterator.getBeginIndex() == iterator.getEndIndex()) {
+                return; /* nothing to draw */
+            }
+            java.awt.font.TextLayout tl = new java.awt.font.TextLayout(iterator, getFontRenderContext());
+            tl.draw(this, (float) x, (float) y);        
         }
 
         public override void drawString(string str, int x, int y)
         {
-            int descent = netfont.FontFamily.GetCellDescent(netfont.Style);
-            int descentPixel = (int)Math.Round(netfont.Size * descent / netfont.FontFamily.GetEmHeight(netfont.Style));
-			g.DrawString(str, netfont, brush, x, y - netfont.Height + descentPixel);
-        }
+			drawString(str, (float)x, (float)y);
+		}
 
-        public override void fill3DRect(int param1, int param2, int param3, int param4, bool param5)
+        public override void fill3DRect(int x, int y, int width, int height, bool raised)
         {
-            throw new NotImplementedException();
+            java.awt.Paint p = getPaint();
+            java.awt.Color c = getColor();
+            java.awt.Color brighter = c.brighter();
+            java.awt.Color darker = c.darker();
+
+            if( !raised ) {
+                setColor(darker);
+            } else if( p != c ) {
+                setColor(c);
+            }
+            fillRect(x + 1, y + 1, width - 2, height - 2);
+            setColor(raised ? brighter : darker);
+            fillRect(x, y, 1, height);
+            fillRect(x + 1, y, width - 2, 1);
+            setColor(raised ? darker : brighter);
+            fillRect(x + 1, y + height - 1, width - 1, 1);
+            fillRect(x + width - 1, y, 1, height - 1);
+            setPaint(p);
         }
 
         public override void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle)
@@ -523,6 +547,9 @@ namespace ikvm.awt
 
         public override void setXORMode(java.awt.Color param)
         {
+            if( param == null ) {
+                throw new java.lang.IllegalArgumentException("null XORColor");
+            }
             throw new NotImplementedException();
         }
 
@@ -541,21 +568,47 @@ namespace ikvm.awt
             }
         }
 
-        public override bool drawImage(java.awt.Image image, java.awt.geom.AffineTransform xform, ImageObserver obs)
+        public override bool drawImage(java.awt.Image img, java.awt.geom.AffineTransform xform, ImageObserver observer)
         {
-            Console.WriteLine(new System.Diagnostics.StackTrace());
-            throw new NotImplementedException();
+            if (img == null) {
+                return true;
+            }
+     
+            if (xform == null || xform.isIdentity()) {
+                return drawImage(img, 0, 0, null, observer);
+            }
+
+            throw new NotImplementedException("drawImage(Image,AffineTransform,ImageObserver) not implemented for non-null or non-identity AffineTransform!");
         }
 
         public override void drawImage(java.awt.image.BufferedImage image, BufferedImageOp op, int x, int y)
         {
-            Console.WriteLine(new System.Diagnostics.StackTrace());
-            throw new NotImplementedException();
+
+            if( op == null ) {
+                drawImage(image, x, y, null);
+            } else {
+                if( !(op is AffineTransformOp) ) {
+                    drawImage(op.filter(image, null), x, y, null);
+                } else {
+                    Console.WriteLine(new System.Diagnostics.StackTrace());
+                    throw new NotImplementedException();
+                }
+            }
         }
 
-        public override void drawRenderedImage(java.awt.image.RenderedImage image, java.awt.geom.AffineTransform xform)
+        public override void drawRenderedImage(java.awt.image.RenderedImage img, java.awt.geom.AffineTransform xform)
         {
-            throw new NotImplementedException();
+            if (img == null) {
+                return;
+            }
+    
+            // BufferedImage case: use a simple drawImage call
+            if (img is BufferedImage) {
+                BufferedImage bufImg = (BufferedImage)img;
+                drawImage(bufImg,xform,null);
+                return;
+            }            
+            throw new NotImplementedException("drawRenderedImage not implemented for images which are not BufferedImages.");
         }
 
         public override void drawRenderableImage(java.awt.image.renderable.RenderableImage image, java.awt.geom.AffineTransform xform)
@@ -565,12 +618,21 @@ namespace ikvm.awt
 
         public override void drawString(string text, float x, float y)
         {
-            g.DrawString(text, netfont, brush, x, y);
-        }
+			float descent = netfont.FontFamily.GetCellDescent(netfont.Style);
+			float descentPixel = netfont.Size / netfont.FontFamily.GetEmHeight(netfont.Style) * descent;
+			g.DrawString(text, netfont, brush, new PointF(x, y - netfont.GetHeight(g) + descentPixel));
+		}
 
         public override void drawString(java.text.AttributedCharacterIterator iterator, float x, float y)
         {
-            throw new NotImplementedException();
+            if( iterator == null ) {
+                throw new java.lang.NullPointerException("AttributedCharacterIterator is null");
+            }
+            if( iterator.getBeginIndex() == iterator.getEndIndex() ) {
+                return; /* nothing to draw */
+            }
+            java.awt.font.TextLayout tl = new java.awt.font.TextLayout(iterator, getFontRenderContext());
+            tl.draw(this, x, y);
         }
 
         public override void fill(java.awt.Shape shape)
@@ -598,7 +660,11 @@ namespace ikvm.awt
 
         public override void setComposite(java.awt.Composite comp)
         {
-            throw new NotImplementedException();
+            if (comp == null)
+            {
+                throw new java.lang.IllegalArgumentException("null Composite");
+            }
+            this.composite = comp;
         }
 
         public override void setPaint(java.awt.Paint paint)
@@ -886,12 +952,15 @@ namespace ikvm.awt
 
         public override java.awt.Paint getPaint()
         {
+            if( javaPaint == null ) {
+                javaPaint = new java.awt.Color(color.ToArgb());
+            }
             return javaPaint;
         }
 
         public override java.awt.Composite getComposite()
         {
-            throw new NotImplementedException();
+            return composite;
         }
 
         public override void setBackground(java.awt.Color color)
@@ -910,17 +979,18 @@ namespace ikvm.awt
             {
                 return defaultStroke;
             }
-            return stroke;
+            return stroke; 
         }
 
         public override java.awt.font.FontRenderContext getFontRenderContext()
         {
-            throw new NotImplementedException();
+            return new java.awt.font.FontRenderContext(getTransform(), false, false);
         }
 
-        public override void drawGlyphVector(java.awt.font.GlyphVector g, float x, float y)
+        public override void drawGlyphVector(java.awt.font.GlyphVector gv, float x, float y)
         {
-            throw new NotImplementedException();
+            char[] text = ((NetGlyphVector)gv).getText();
+            drawString(new string(text), x, y);
         }
     }
 
@@ -950,7 +1020,7 @@ namespace ikvm.awt
 
         public override java.awt.image.BufferedImage createCompatibleImage(int width, int height)
         {
-            return new NetBufferedImage(width, height);
+            return new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
         }
 
         public override java.awt.image.VolatileImage createCompatibleVolatileImage(int param1, int param2, java.awt.ImageCapabilities param3)
