@@ -21,14 +21,24 @@
   jeroen@frijters.net
   
 */
+#if IKVM_REF_EMIT
+// FXBUG multi target only work with our own Reflection.Emit implementation
+#define MULTI_TARGET
+#endif
+
 using System;
 using System.Collections.Generic;
 using System.IO;
 using System.Reflection;
+#if IKVM_REF_EMIT
+using IKVM.Reflection.Emit;
+#else
 using System.Reflection.Emit;
+#endif
 using System.Threading;
 using ICSharpCode.SharpZipLib.Zip;
 using IKVM.Internal;
+using System.Text.RegularExpressions;
 
 class IkvmcCompiler
 {
@@ -55,7 +65,7 @@ class IkvmcCompiler
 					string line;
 					while((line = sr.ReadLine()) != null)
 					{
-						arglist.Add(line);
+						arglist.Add(line.Trim());
 					}
 				}
 			}
@@ -72,6 +82,7 @@ class IkvmcCompiler
 		DateTime start = DateTime.Now;
 		AppDomain.CurrentDomain.ReflectionOnlyAssemblyResolve += new ResolveEventHandler(CurrentDomain_AssemblyResolve);
 		System.Threading.Thread.CurrentThread.Name = "compiler";
+		Tracer.EnableTraceConsoleListener();
 		Tracer.EnableTraceForDebug();
 		List<string> argList = GetArgs(args);
 		if (argList.Count == 0)
@@ -81,7 +92,7 @@ class IkvmcCompiler
 		}
 		IkvmcCompiler comp = new IkvmcCompiler();
 		List<CompilerOptions> targets = new List<CompilerOptions>();
-		int rc = comp.ParseCommandLine(argList.GetEnumerator(), targets, new CompilerOptions());
+		int rc = comp.ParseCommandLine(argList.GetEnumerator(), targets);
 		if (rc == 0)
 		{
 			try
@@ -105,7 +116,13 @@ class IkvmcCompiler
 		if (time)
 		{
 			Console.WriteLine("Total cpu time: {0}", System.Diagnostics.Process.GetCurrentProcess().TotalProcessorTime);
+			Console.WriteLine("User cpu time: {0}", System.Diagnostics.Process.GetCurrentProcess().UserProcessorTime);
 			Console.WriteLine("Total wall clock time: {0}", DateTime.Now - start);
+			Console.WriteLine("Peak virtual memory: {0}", System.Diagnostics.Process.GetCurrentProcess().PeakVirtualMemorySize64);
+			for (int i = 0; i <= GC.MaxGeneration; i++)
+			{
+				Console.WriteLine("GC({0}) count: {1}", i, GC.CollectionCount(i));
+			}
 		}
 		return rc;
 	}
@@ -186,15 +203,25 @@ class IkvmcCompiler
 		Console.Error.WriteLine("    -warnaserror:<warning[:key]>  Treat specified warnings as errors");
 		Console.Error.WriteLine("    -time                      Display timing statistics");
 		Console.Error.WriteLine("    -classloader:<class>       Set custom class loader class for assembly");
+#if MULTI_TARGET
+		Console.Error.WriteLine("    -sharedclassloader         All targets below this level share a common");
+		Console.Error.WriteLine("                               class loader");
+#endif
 	}
 
-	int ParseCommandLine(IEnumerator<string> arglist, List<CompilerOptions> targets, CompilerOptions options)
+	int ParseCommandLine(IEnumerator<string> arglist, List<CompilerOptions> targets)
 	{
+		CompilerOptions options = new CompilerOptions();
 		options.target = PEFileKinds.ConsoleApplication;
 		options.guessFileKind = true;
 		options.version = "0.0.0.0";
 		options.apartment = ApartmentState.STA;
 		options.props = new Dictionary<string, string>();
+		return ContinueParseCommandLine(arglist, targets, options);
+	}
+
+	int ContinueParseCommandLine(IEnumerator<string> arglist, List<CompilerOptions> targets, CompilerOptions options)
+	{
 		while(arglist.MoveNext())
 		{
 			string s = arglist.Current;
@@ -209,7 +236,7 @@ class IkvmcCompiler
 				nestedLevel.defaultAssemblyName = defaultAssemblyName;
 				nestedLevel.classesToExclude = new List<string>(classesToExclude);
 				nestedLevel.references = new List<string>(references);
-				int rc = nestedLevel.ParseCommandLine(arglist, targets, options.Copy());
+				int rc = nestedLevel.ContinueParseCommandLine(arglist, targets, options.Copy());
 				if(rc != 0)
 				{
 					return rc;
@@ -417,7 +444,14 @@ class IkvmcCompiler
 						try
 						{
 							DirectoryInfo dir = new DirectoryInfo(Path.GetDirectoryName(spec));
-							Recurse(dir, dir, Path.GetFileName(spec));
+							if(dir.Exists)
+							{
+								Recurse(dir, dir, Path.GetFileName(spec));
+							}
+							else
+							{
+								RecurseJar(spec);
+							}
 						}
 						catch(PathTooLongException)
 						{
@@ -610,6 +644,15 @@ class IkvmcCompiler
 				{
 					options.classLoader = s.Substring(13);
 				}
+#if MULTI_TARGET
+				else if(s == "-sharedclassloader")
+				{
+					if(options.sharedclassloader == null)
+					{
+						options.sharedclassloader = new List<CompilerClassLoader>();
+					}
+				}
+#endif
 				else
 				{
 					Console.Error.WriteLine("Warning: unrecognized option: {0}", s);
@@ -650,6 +693,11 @@ class IkvmcCompiler
 					ProcessFile(null, f);
 				}
 			}
+			if(options.targetIsModule && options.sharedclassloader != null)
+			{
+				Console.Error.WriteLine("Error: -target:module and -sharedclassloader options cannot be combined.");
+				return 1;
+			}
 		}
 #if MULTI_TARGET
 		if(nonleaf)
@@ -738,7 +786,7 @@ class IkvmcCompiler
 		}
 	}
 
-	private void ProcessZipFile(string file)
+	private void ProcessZipFile(string file, Predicate<ZipEntry> filter)
 	{
 		ZipFile zf = new ZipFile(file);
 		try
@@ -749,6 +797,10 @@ class IkvmcCompiler
 				{
 					// skip
 				}
+				else if(filter != null && !filter(ze))
+				{
+					// skip
+				}
 				else if(ze.Name.ToLower().EndsWith(".class"))
 				{
 					AddClassFile(ze.Name, ReadFromZip(zf, ze), true);
@@ -810,7 +862,7 @@ class IkvmcCompiler
 			case ".zip":
 				try
 				{
-					ProcessZipFile(file);
+					ProcessZipFile(file, null);
 				}
 				catch(ICSharpCode.SharpZipLib.SharpZipBaseException x)
 				{
@@ -864,6 +916,30 @@ class IkvmcCompiler
 		}
 	}
 
+	private void RecurseJar(string path)
+	{
+		string file = "";
+		for (; ; )
+		{
+			file = Path.Combine(Path.GetFileName(path), file);
+			path = Path.GetDirectoryName(path);
+			if (Directory.Exists(path))
+			{
+				throw new DirectoryNotFoundException();
+			}
+			else if (File.Exists(path))
+			{
+				string pathFilter = Path.GetDirectoryName(file) + Path.DirectorySeparatorChar;
+				string fileFilter = "^" + Regex.Escape(Path.GetFileName(file)).Replace("\\*", ".*").Replace("\\?", ".") + "$";
+				ProcessZipFile(path, delegate(ZipEntry ze) {
+					return (Path.GetDirectoryName(ze.Name) + Path.DirectorySeparatorChar).StartsWith(pathFilter)
+						&& Regex.IsMatch(Path.GetFileName(ze.Name), fileFilter);
+				});
+				return;
+			}
+		}
+	}
+
 	//This processes an exclusion file with a single regular expression per line
 	private static void ProcessExclusionFile(List<string> classesToExclude, String filename)
 	{
