@@ -26,7 +26,11 @@ using System;
 using System.Collections.Generic;
 using System.Xml.Serialization;
 using System.Reflection;
+#if IKVM_REF_EMIT
+using IKVM.Reflection.Emit;
+#else
 using System.Reflection.Emit;
+#endif
 using System.Diagnostics;
 using IKVM.Attributes;
 using IKVM.Internal;
@@ -358,6 +362,22 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("castclass_impl")]
+	public sealed class Castclass_impl : Instruction
+	{
+		[XmlAttribute("class")]
+		public string Class;
+
+		public Castclass_impl()
+		{
+		}
+
+		internal override void Generate(CodeGenContext context, CodeEmitter ilgen)
+		{
+			ilgen.Emit(OpCodes.Castclass, context.ClassLoader.LoadClassByDottedName(Class).TypeAsBaseType);
+		}
+	}
+
 	public abstract class TypeInstruction : Instruction
 	{
 		[XmlAttribute("type")]
@@ -1085,6 +1105,18 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("newarr")]
+	public sealed class Newarr : Instruction
+	{
+		[XmlAttribute("sig")]
+		public string Sig;
+
+		internal override void Generate(CodeGenContext context, CodeEmitter ilgen)
+		{
+			ilgen.Emit(OpCodes.Newarr, context.ClassLoader.FieldTypeWrapperFromSig(Sig).TypeAsArrayType);
+		}
+	}
+
 	[XmlType("ldtoken")]
 	public sealed class Ldtoken : Instruction
 	{
@@ -1130,6 +1162,7 @@ namespace IKVM.Internal.MapXml
 		[XmlElement(typeof(Pop))]
 		[XmlElement(typeof(IsInst))]
 		[XmlElement(typeof(Castclass))]
+		[XmlElement(typeof(Castclass_impl))]
 		[XmlElement(typeof(Ldobj))]
 		[XmlElement(typeof(Unbox))]
 		[XmlElement(typeof(Box))]
@@ -1194,6 +1227,7 @@ namespace IKVM.Internal.MapXml
 		[XmlElement(typeof(ConditionalInstruction))]
 		[XmlElement(typeof(Volatile))]
 		[XmlElement(typeof(Ldelema))]
+		[XmlElement(typeof(Newarr))]
 		[XmlElement(typeof(Ldtoken))]
 		[XmlElement(typeof(Leave))]
 		[XmlElement(typeof(RunClassInit))]
