@@ -25,9 +25,14 @@
 using System;
 using System.Collections.Generic;
 using System.Reflection;
+#if IKVM_REF_EMIT
+using IKVM.Reflection.Emit;
+#else
 using System.Reflection.Emit;
+#endif
 using System.Runtime.InteropServices;
 using System.Diagnostics.SymbolStore;
+using System.Diagnostics;
 
 namespace IKVM.Internal
 {
@@ -71,7 +76,8 @@ namespace IKVM.Internal
 #if STATIC_COMPILER
 		private IKVM.Attributes.LineNumberTableAttribute.LineNumberWriter linenums;
 #endif // STATIC_COMPILER
-		private Expr stack;
+		private Expr[] stackArray = new Expr[8];
+		private int topOfStack;
 		private CodeEmitterLabel lazyBranch;
 		private LocalBuilder[] tempLocals = new LocalBuilder[32];
 #if LABELCHECK
@@ -100,6 +106,42 @@ namespace IKVM.Internal
 			this.ilgen_real = ilgen;
 		}
 
+		private void PushStack(Expr expr)
+		{
+			Debug.Assert(expr != null);
+			if (topOfStack == stackArray.Length)
+			{
+				Array.Resize(ref stackArray, stackArray.Length * 2);
+			}
+			stackArray[topOfStack++] = expr;
+		}
+
+		private void PushStackMayBeNull(Expr expr)
+		{
+			if (expr != null)
+			{
+				PushStack(expr);
+			}
+		}
+
+		private Expr PopStack()
+		{
+			if (topOfStack == 0)
+			{
+				return null;
+			}
+			return stackArray[--topOfStack];
+		}
+
+		private Expr PeekStack()
+		{
+			if (topOfStack == 0)
+			{
+				return null;
+			}
+			return stackArray[topOfStack - 1];
+		}
+
 		internal LocalBuilder UnsafeAllocTempLocal(Type type)
 		{
 			int free = -1;
@@ -156,7 +198,7 @@ namespace IKVM.Internal
 		{
 			get
 			{
-				return stack == null;
+				return topOfStack == 0;
 			}
 		}
 
@@ -536,45 +578,66 @@ namespace IKVM.Internal
 		}
 #endif // STATIC_COMPILER
 
+		internal void LazyEmitPop()
+		{
+			Expr exp = PeekStack();
+			if (exp == null || exp.HasSideEffect || exp.IsIncomplete)
+			{
+				Emit(OpCodes.Pop);
+			}
+			else
+			{
+				PopStack();
+			}
+		}
+
+		internal void LazyEmitLoadClass(TypeWrapper type)
+		{
+			PushStack(new ClassLiteralExpr(type));
+		}
+
 		internal void LazyEmitBox(Type type)
 		{
-			stack = new BoxExpr(stack, type);
+			PushStack(new BoxExpr(PopStack(), type));
 		}
 
 		internal void LazyEmitUnbox(Type type)
 		{
-			BoxExpr box = stack as BoxExpr;
+			BoxExpr box = PeekStack() as BoxExpr;
 			if(box != null)
 			{
-				stack = new BoxUnboxExpr(box.Expr, type);
+				PopStack();
+				PushStack(new BoxUnboxExpr(box.Expr, type));
 			}
 			else
 			{
-				stack = new UnboxExpr(stack, type);
+				PushStack(new UnboxExpr(PopStack(), type));
 			}
 		}
 
 		internal void LazyEmitLdobj(Type type)
 		{
-			BoxUnboxExpr boxunbox = stack as BoxUnboxExpr;
+			BoxUnboxExpr boxunbox = PeekStack() as BoxUnboxExpr;
 			if(boxunbox != null)
 			{
+				PopStack();
 				// box/unbox+ldobj annihilate each other
-				stack = boxunbox.Expr;
+				PushStackMayBeNull(boxunbox.Expr);
 			}
 			else
 			{
-				stack = new LdobjExpr(stack, type);
+				PushStack(new LdobjExpr(PopStack(), type));
 			}
 		}
 
 		internal void LazyEmitUnboxSpecial(Type type)
 		{
-			BoxExpr box = stack as BoxExpr;
+			BoxExpr box = PeekStack() as BoxExpr;
 			if(box != null)
 			{
+				PopStack();
 				// the unbox and lazy box cancel each other out
-				stack = box.Expr;
+				PushStackMayBeNull(box.Expr);
 			}
 			else
 			{
@@ -593,22 +656,24 @@ namespace IKVM.Internal
 			}
 		}
 
+		internal void LazyEmitLdnull()
+		{
+			PushStack(new NullExpr());
+		}
+
 		internal void LazyEmitLdc_I4(int i)
 		{
-			LazyGen();
-			stack = new ConstIntExpr(i);
+			PushStack(new ConstIntExpr(i));
 		}
 
 		internal void LazyEmitLdc_I8(long l)
 		{
-			LazyGen();
-			stack = new ConstLongExpr(l);
+			PushStack(new ConstLongExpr(l));
 		}
 
 		internal void LazyEmitLdstr(string str)
 		{
-			LazyGen();
-			stack = new ConstStringExpr(str);
+			PushStack(new ConstStringExpr(str));
 		}
 
 		internal void LazyEmit_idiv()
@@ -616,12 +681,12 @@ namespace IKVM.Internal
 			// we need to special case dividing by -1, because the CLR div instruction
 			// throws an OverflowException when dividing Int32.MinValue by -1, and
 			// Java just silently overflows
-			ConstIntExpr v = stack as ConstIntExpr;
+			ConstIntExpr v = PeekStack() as ConstIntExpr;
 			if(v != null)
 			{
 				if(v.i == -1)
 				{
-					stack = null;
+					PopStack();
 					Emit(OpCodes.Neg);
 				}
 				else
@@ -650,12 +715,12 @@ namespace IKVM.Internal
 			// we need to special case dividing by -1, because the CLR div instruction
 			// throws an OverflowException when dividing Int32.MinValue by -1, and
 			// Java just silently overflows
-			ConstLongExpr v = stack as ConstLongExpr;
+			ConstLongExpr v = PeekStack() as ConstLongExpr;
 			if(v != null)
 			{
 				if(v.l == -1)
 				{
-					stack = null;
+					PopStack();
 					Emit(OpCodes.Neg);
 				}
 				else
@@ -682,8 +747,7 @@ namespace IKVM.Internal
 
 		internal void LazyEmit_instanceof(Type type)
 		{
-			LazyGen();
-			stack = new InstanceOfExpr(type);
+			PushStack(new InstanceOfExpr(type));
 		}
 
 		internal void LazyEmit_ifeq(CodeEmitterLabel label)
@@ -698,18 +762,18 @@ namespace IKVM.Internal
 
 		private void LazyEmit_if_ne_eq(CodeEmitterLabel label, bool brtrue)
 		{
-			InstanceOfExpr instanceof = stack as InstanceOfExpr;
+			InstanceOfExpr instanceof = PeekStack() as InstanceOfExpr;
 			if (instanceof != null)
 			{
-				stack = null;
+				PopStack();
 				Emit(OpCodes.Isinst, instanceof.Type);
 			}
 			else
 			{
-				CmpExpr cmp = stack as CmpExpr;
+				CmpExpr cmp = PeekStack() as CmpExpr;
 				if (cmp != null)
 				{
-					stack = null;
+					PopStack();
 					Emit(brtrue ? OpCodes.Bne_Un : OpCodes.Beq, label);
 					return;
 				}
@@ -746,10 +810,10 @@ namespace IKVM.Internal
 
 		internal void LazyEmit_if_le_lt_ge_gt(Comparison comp, CodeEmitterLabel label)
 		{
-			CmpExpr cmp = stack as CmpExpr;
+			CmpExpr cmp = PeekStack() as CmpExpr;
 			if (cmp != null)
 			{
-				stack = null;
+				PopStack();
 				cmp.EmitBcc(this, comp, label);
 			}
 			else
@@ -761,40 +825,35 @@ namespace IKVM.Internal
 
 		internal void LazyEmit_lcmp()
 		{
-			LazyGen();
-			stack = new LCmpExpr();
+			PushStack(new LCmpExpr());
 		}
 
 		internal void LazyEmit_fcmpl()
 		{
-			LazyGen();
-			stack = new FCmplExpr();
+			PushStack(new FCmplExpr());
 		}
 
 		internal void LazyEmit_fcmpg()
 		{
-			LazyGen();
-			stack = new FCmpgExpr();
+			PushStack(new FCmpgExpr());
 		}
 
 		internal void LazyEmit_dcmpl()
 		{
-			LazyGen();
-			stack = new DCmplExpr();
+			PushStack(new DCmplExpr());
 		}
 
 		internal void LazyEmit_dcmpg()
 		{
-			LazyGen();
-			stack = new DCmpgExpr();
+			PushStack(new DCmpgExpr());
 		}
 
 		internal void LazyEmitAnd_I4(int v2)
 		{
-			ConstIntExpr v1 = stack as ConstIntExpr;
+			ConstIntExpr v1 = PeekStack() as ConstIntExpr;
 			if (v1 != null)
 			{
-				stack = null;
+				PopStack();
 				LazyEmitLdc_I4(v1.i & v2);
 			}
 			else
@@ -806,29 +865,39 @@ namespace IKVM.Internal
 
 		internal string PopLazyLdstr()
 		{
-			ConstStringExpr str = stack as ConstStringExpr;
+			ConstStringExpr str = PeekStack() as ConstStringExpr;
 			if(str != null)
 			{
-				stack = null;
+				PopStack();
 				return str.str;
 			}
 			return null;
 		}
 
+		internal TypeWrapper PeekLazyClassLiteral()
+		{
+			ClassLiteralExpr lit = PeekStack() as ClassLiteralExpr;
+			if (lit != null)
+			{
+				return lit.Type;
+			}
+			return null;
+		}
+
 		private void LazyGen()
 		{
-			if(stack != null)
-			{
-				Expr exp = stack;
-				stack = null;
-				exp.Emit(this);
-			}
 			if(lazyBranch != null)
 			{
 				offset += OpCodes.Br.Size + 4;
 				ilgen_real.Emit(OpCodes.Br, lazyBranch.Label);
 				lazyBranch = null;
 			}
+			int len = topOfStack;
+			topOfStack = 0;
+			for(int i = 0; i < len; i++)
+			{
+				stackArray[i].Emit(this);
+			}
 		}
 
 		internal void Finish()
@@ -844,24 +913,27 @@ namespace IKVM.Internal
 
 		abstract class Expr
 		{
-			internal readonly Type Type;
+			internal bool HasSideEffect { get { return false; } }	// for now we only have side-effect free expressions
 
-			protected Expr(Type type)
-			{
-				this.Type = type;
-			}
+			internal virtual bool IsIncomplete { get { return false; } }
 
 			internal abstract void Emit(CodeEmitter ilgen);
 		}
 
-		abstract class UnaryExpr : Expr
+		abstract class ExprWithExprAndType : Expr
 		{
 			internal readonly Expr Expr;
+			internal readonly Type Type;
 
-			protected UnaryExpr(Expr expr, Type type)
-				: base(type)
+			protected ExprWithExprAndType(Expr expr, Type type)
 			{
 				this.Expr = expr;
+				this.Type = type;
+			}
+
+			internal override bool IsIncomplete
+			{
+				get { return Expr == null; }
 			}
 
 			internal override void Emit(CodeEmitter ilgen)
@@ -873,7 +945,7 @@ namespace IKVM.Internal
 			}
 		}
 
-		class BoxExpr : UnaryExpr
+		class BoxExpr : ExprWithExprAndType
 		{
 			internal BoxExpr(Expr expr, Type type)
 				: base(expr, type)
@@ -887,7 +959,7 @@ namespace IKVM.Internal
 			}
 		}
 
-		class UnboxExpr : UnaryExpr
+		class UnboxExpr : ExprWithExprAndType
 		{
 			internal UnboxExpr(Expr expr, Type type)
 				: base(expr, type)
@@ -901,7 +973,7 @@ namespace IKVM.Internal
 			}
 		}
 
-		class BoxUnboxExpr : UnaryExpr
+		class BoxUnboxExpr : ExprWithExprAndType
 		{
 			internal BoxUnboxExpr(Expr expr, Type type)
 				: base(expr, type)
@@ -920,7 +992,7 @@ namespace IKVM.Internal
 			}
 		}
 
-		class LdobjExpr : UnaryExpr
+		class LdobjExpr : ExprWithExprAndType
 		{
 			internal LdobjExpr(Expr expr, Type type)
 				: base(expr, type)
@@ -934,12 +1006,19 @@ namespace IKVM.Internal
 			}
 		}
 
+		class NullExpr : Expr
+		{
+			internal override void Emit(CodeEmitter ilgen)
+			{
+				ilgen.Emit(OpCodes.Ldnull);
+			}
+		}
+
 		class ConstIntExpr : Expr
 		{
 			internal readonly int i;
 
 			internal ConstIntExpr(int i)
-				: base(typeof(int))
 			{
 				this.i = i;
 			}
@@ -997,7 +1076,6 @@ namespace IKVM.Internal
 			internal readonly long l;
 
 			internal ConstLongExpr(long l)
-				: base(typeof(long))
 			{
 				this.l = l;
 			}
@@ -1080,7 +1158,6 @@ namespace IKVM.Internal
 			internal readonly string str;
 
 			internal ConstStringExpr(string str)
-				: base(typeof(string))
 			{
 				this.str = str;
 			}
@@ -1093,9 +1170,19 @@ namespace IKVM.Internal
 
 		sealed class InstanceOfExpr : Expr
 		{
+			internal readonly Type Type;
+
 			internal InstanceOfExpr(Type type)
-				: base(type)
 			{
+				this.Type = type;
+			}
+
+			internal override bool IsIncomplete
+			{
+				get
+				{
+					return true;
+				}
 			}
 
 			internal override void Emit(CodeEmitter ilgen)
@@ -1109,10 +1196,17 @@ namespace IKVM.Internal
 		abstract class CmpExpr : Expr
 		{
 			internal CmpExpr()
-				: base(typeof(int))
 			{
 			}
 
+			internal override bool IsIncomplete
+			{
+				get
+				{
+					return true;
+				}
+			}
+
 			internal abstract void EmitBcc(CodeEmitter ilgen, Comparison comp, CodeEmitterLabel label);
 		}
 
@@ -1248,6 +1342,44 @@ namespace IKVM.Internal
 				return typeof(double);
 			}
 		}
+
+		class ClassLiteralExpr : Expr
+		{
+			internal readonly TypeWrapper Type;
+
+			internal ClassLiteralExpr(TypeWrapper type)
+			{
+				this.Type = type;
+			}
+
+			internal override void Emit(CodeEmitter ilgen)
+			{
+				TypeWrapper tw = Type;
+				if (tw.IsGhostArray)
+				{
+					int rank = tw.ArrayRank;
+					while (tw.IsArray)
+					{
+						tw = tw.ElementTypeWrapper;
+					}
+					ilgen.Emit(OpCodes.Ldsfld, RuntimeHelperTypes.GetClassLiteralField(ArrayTypeWrapper.MakeArrayType(tw.TypeAsTBD, rank)));
+				}
+				else
+				{
+					Type type = tw.IsRemapped ? tw.TypeAsBaseType : tw.TypeAsTBD;
+					if (type == typeof(void))
+					{
+						// the CLR doesn't allow System.Void to be used as a type parameter, so we have to special case it
+						ilgen.Emit(OpCodes.Ldtoken, type);
+						Compiler.getClassFromTypeHandle.EmitCall(ilgen);
+					}
+					else
+					{
+						ilgen.Emit(OpCodes.Ldsfld, RuntimeHelperTypes.GetClassLiteralField(type));
+					}
+				}
+			}
+		}
 	}
 }
 #endif
